from __future__ import annotations

from dataclasses import dataclass
from typing import Optional


class LamenError(Exception):
    """Base exception for the SDK."""


@dataclass
class ApiError(LamenError):
    status_code: int
    message: str
    method: str
    path: str
    response_json: Optional[dict] = None
    request_id: Optional[str] = None

    def __str__(self) -> str:
        rid = f" request_id={self.request_id}" if self.request_id else ""
        return f"ApiError({self.status_code}) {self.method} {self.path}: {self.message}{rid}"


class AuthenticationError(LamenError):
    pass


class BillingError(LamenError):
    pass


class ValidationError(LamenError):
    pass


class RateLimitError(LamenError):
    pass


class TimeoutError(LamenError):
    pass


class NetworkError(LamenError):
    pass



from __future__ import annotations

import re
from datetime import datetime
from typing import Any, Dict, Optional
from uuid import UUID

from pydantic import BaseModel, EmailStr, Field, field_validator, ConfigDict

MAX_EVENT_NAME = 80
MAX_EXTERNAL_USER_ID = 128
MAX_IDEMPOTENCY_KEY = 80


# ============================
# STORAGE
# ============================

def _normalize_folder_path(v: str) -> str:
    """Normalize and validate folder paths.

    Mirrors backend behaviour:
    - trim whitespace + leading/trailing slashes
    - convert whitespace runs to a single underscore
    - reject empty paths
    - reject '.' to prevent traversal-like paths
    """
    v = (v or "").strip().strip("/")
    v = re.sub(r"\s+", "_", v)

    if not v:
        raise ValueError("Folder path cannot be empty")
    if "." in v:
        raise ValueError("Invalid folder path")
    return v


class StorageUploadRequest(BaseModel):
    filename: str
    content_type: str
    size_bytes: int
    is_public: bool = False
    folder_id: Optional[UUID] = None
    generate_thumbnail: bool = False
    transcode_preset: str = Field(default="none", max_length=20)
    
    @field_validator("size_bytes")
    @classmethod
    def validate_size_bytes(cls, v: int) -> int:
        iv = int(v)
        if iv <= 0:
            raise ValueError("File size must be positive")
        return iv
    
    @field_validator("transcode_preset")
    @classmethod
    def validate_transcode_preset(cls, v: str) -> str:
        v = (v or "none").strip().lower()
        if v not in {"none", "normalize"}:
            raise ValueError("Invalid transcode_preset")
        return v


class StorageConfirmRequest(BaseModel):
    object_id: UUID


class StorageDownloadRequest(BaseModel):
    object_id: UUID


class StoragePrivacyRequest(BaseModel):
    is_public: bool


class FolderCreateRequest(BaseModel):
    path: str

    @field_validator("path")
    @classmethod
    def validate_path(cls, v: str) -> str:
        return _normalize_folder_path(v)


# ----------------------------
# Responses
# ----------------------------

class StorageUploadResponse(BaseModel):
    uploadUrl: str
    fileUrl: str
    object_id: UUID


class StorageDownloadResponse(BaseModel):
    downloadUrl: str
    expiresIn: int


class StorageObjectOut(BaseModel):
    id: UUID
    path: str
    size_bytes: Optional[int] = None
    content_type: Optional[str] = None
    is_public: bool
    created_at: datetime
    folder_id: Optional[UUID] = None
    status: str
    expires_at: Optional[datetime] = None
    thumbnail_object_id: Optional[UUID] = None
    transcode_preset: str = "none"
    transcode_status: str = "none"
    transcode_failure_reason: Optional[str] = None
    normalized_object_id: Optional[UUID] = None
    model_config = ConfigDict(from_attributes=True)



class StorageDeleteResponse(BaseModel):
    detail: str


class FolderOut(BaseModel):
    id: UUID
    path: str
    created_at: datetime

    class Config:
        from_attributes = True


class FolderDeleteResponse(BaseModel):
    detail: str


# ============================
# STREAMING / CDN
# ============================

class StreamUrlResponse(BaseModel):
    url: str
    expires_in: int


# ============================
# ATHENA
# ============================

class AthenaLogicEventIngest(BaseModel):
    external_user_id: str
    event_name: str
    occurred_at: datetime
    properties: Optional[Dict[str, Any]] = None
    idempotency_key: Optional[str] = Field(default=None, max_length=255)


class AthenaIngestResponse(BaseModel):
    ok: bool
    deduplicated: Optional[bool] = None


# ============================
# MESSAGING (API key)
# ============================

class EventIngestRequest(BaseModel):
    event: str = Field(..., min_length=1, max_length=MAX_EVENT_NAME)
    external_user_id: Optional[str] = Field(None, min_length=1, max_length=MAX_EXTERNAL_USER_ID)
    data: Optional[Dict[str, Any]] = None
    idempotency_key: Optional[str] = Field(None, max_length=MAX_IDEMPOTENCY_KEY)


class EventIngestResponse(BaseModel):
    event_id: UUID
    status: str
    deliveries_planned: int


class RegisterDeviceIdentityRequest(BaseModel):
    external_user_id: str = Field(..., min_length=1, max_length=MAX_EXTERNAL_USER_ID)
    token: str = Field(..., min_length=10, max_length=512)
    platform: str = Field(..., min_length=2, max_length=16)
    app_id: Optional[str] = Field(None, max_length=64)


class DeviceIdentityResponse(BaseModel):
    id: UUID
    external_user_id: str
    platform: str
    app_id: Optional[str]
    is_active: bool
    created_at: datetime
    last_seen_at: Optional[datetime] = None


class UnregisterDeviceIdentityRequest(BaseModel):
    external_user_id: str = Field(..., min_length=1, max_length=MAX_EXTERNAL_USER_ID)
    token: str = Field(..., min_length=10, max_length=512)


class UnregisterDeviceIdentityResponse(BaseModel):
    ok: bool


class RegisterEmailIdentityRequest(BaseModel):
    external_user_id: str = Field(..., min_length=1, max_length=MAX_EXTERNAL_USER_ID)
    email: str = Field(..., min_length=3, max_length=320)
    label: Optional[str] = Field(None, max_length=64)


class EmailIdentityResponse(BaseModel):
    id: UUID
    external_user_id: str
    email: str
    label: Optional[str]
    verification_level: str    # "none" | "user_verified" | "provider_verified" | "workspace_verified"
    is_active: bool
    created_at: datetime
    last_seen_at: Optional[datetime]


class StartEmailVerifyRequest(BaseModel):
    external_user_id: str = Field(..., min_length=1, max_length=MAX_EXTERNAL_USER_ID)
    email: EmailStr
    label: Optional[str] = Field(None, max_length=64)


class StartEmailVerifyResponse(BaseModel):
    ok: bool


class ConfirmEmailVerifyRequest(BaseModel):
    external_user_id: str = Field(..., min_length=1, max_length=MAX_EXTERNAL_USER_ID)
    email: EmailStr
    code: str = Field(..., min_length=4, max_length=12)


class ConfirmEmailVerifyResponse(BaseModel):
    ok: bool


class WebTokenRegisterRequest(BaseModel):
    external_user_id: str = Field(..., min_length=1, max_length=MAX_EXTERNAL_USER_ID)
    token: str = Field(..., min_length=1, max_length=2048)
    app_id: Optional[str] = Field(None, max_length=64)


class WebTokenUnregisterRequest(BaseModel):
    external_user_id: str = Field(..., min_length=1, max_length=MAX_EXTERNAL_USER_ID)
    token: str = Field(..., min_length=1, max_length=2048)


class DirectSendRequest(BaseModel):
    channel: str
    external_user_id: str
    title: Optional[str] = None
    body: Optional[str] = None
    data: Optional[Dict[str, Any]] = None
    priority: str = "standard"
    idempotency_key: Optional[str] = None


class DirectSendResponse(BaseModel):
    ok: bool
    deliveries_created: int
    
class BulkEmailImportRequest(BaseModel):
    users: list[dict]
    

class GeocodeResult(BaseModel):
    id: str
    label: str
    lat: float
    lng: float
    type: str = "unknown"     
    source: str = "lamen-osm"


from __future__ import annotations

from datetime import datetime
from typing import Any, Dict, Optional
from .http import HttpClient, AsyncHttpClient
from .types import AthenaLogicEventIngest, AthenaIngestResponse


class AthenaClient:
    """Athena ingest (API key): POST /athena/api/events"""

    def __init__(self, http: HttpClient):
        self._http = http

    def ingest_event(
        self,
        *,
        external_user_id: str,
        event_name: str,
        occurred_at: datetime,
        properties: Optional[Dict[str, Any]] = None,
        idempotency_key: Optional[str] = None,
    ) -> AthenaIngestResponse:
        payload = AthenaLogicEventIngest(
            external_user_id=external_user_id,
            event_name=event_name,
            occurred_at=occurred_at,
            properties=properties,
            idempotency_key=idempotency_key,
        )
        data = self._http.post("/athena/api/events", json=payload.model_dump())
        return AthenaIngestResponse.model_validate(data)


class AsyncAthenaClient:
    def __init__(self, http: AsyncHttpClient):
        self._http = http

    async def ingest_event(
        self,
        *,
        external_user_id: str,
        event_name: str,
        occurred_at: datetime,
        properties: Optional[Dict[str, Any]] = None,
        idempotency_key: Optional[str] = None,
    ) -> AthenaIngestResponse:
        payload = AthenaLogicEventIngest(
            external_user_id=external_user_id,
            event_name=event_name,
            occurred_at=occurred_at,
            properties=properties,
            idempotency_key=idempotency_key,
        )
        data = await self._http.post("/athena/api/events", json=payload.model_dump())
        return AthenaIngestResponse.model_validate(data)



from __future__ import annotations

import asyncio
import random
from dataclasses import dataclass
from typing import Any, Dict, Optional, Callable
import logging
from ._version import __version__

import httpx

from .errors import (
    ApiError,
    AuthenticationError,
    BillingError,
    NetworkError,
    RateLimitError,
    TimeoutError,
    ValidationError,
)

DEFAULT_BASE_URL = "https://api.lamen.dev"
DEFAULT_TIMEOUT_S = 30.0

def _redact_headers(headers: dict[str, str]) -> dict[str, str]:
    redacted = {}
    for k, v in headers.items():
        kl = k.lower()
        if kl in {"authorization", "x-api-key"}:
            redacted[k] = "***"
        else:
            redacted[k] = v
    return redacted


@dataclass(frozen=True)
class RetryConfig:
    """Retry policy for transient failures.

    Notes:
      - Retries are only safe for *idempotent* operations.
      - For POST requests, you should provide an idempotency key (either a field
        in the payload, or a header on endpoints that support it).
    """

    max_retries: int = 2
    base_backoff_s: float = 0.25
    max_backoff_s: float = 4.0
    jitter_s: float = 0.15


class HttpClient:
    """Low-level sync HTTP wrapper with retries and typed errors."""

    def __init__(
        self,
        *,
        api_key: Optional[str] = None,
        base_url: str = DEFAULT_BASE_URL,
        dashboard_token: Optional[str] = None,
        timeout_s: float = DEFAULT_TIMEOUT_S,
        user_agent: str = f"lamen-python/{__version__}",
        retry: RetryConfig | None = None,
        extra_headers: Optional[Dict[str, str]] = None,
        logger: logging.Logger | None = None,
        debug: bool = False,
    ) -> None:
        self._retry = retry or RetryConfig()
        self._logger = logger or logging.getLogger("lamen")
        self._debug = debug

        headers: Dict[str, str] = {
            "User-Agent": user_agent,
            "Accept": "application/json",
        }
        if api_key:
            headers["x-api-key"] = api_key
        if dashboard_token:
            headers["Authorization"] = f"Bearer {dashboard_token}"
        if extra_headers:
            headers.update({k: v for k, v in extra_headers.items() if v is not None})

        self._client = httpx.Client(
            base_url=base_url.rstrip("/"),
            headers=headers,
            timeout=httpx.Timeout(timeout_s),
        )
        
    def _has_idempotency_key(
        self,
        *,
        json: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, str]] = None,
    ) -> bool:
        # Header-based idempotency
        if headers:
            for k in ("Idempotency-Key", "X-Idempotency-Key", "x-idempotency-key"):
                v = headers.get(k)
                if v is not None and str(v).strip():
                    return True

        # Body-based idempotency (preferred in Lamen)
        if json and isinstance(json, dict):
            v = json.get("idempotency_key")
            if v is not None and str(v).strip():
                return True

        return False

    def _effective_attempts(
        self,
        *,
        method_u: str,
        json: Optional[Dict[str, Any]],
        headers: Optional[Dict[str, str]],
        cfg: RetryConfig,
    ) -> int:
        # Safe methods
        if method_u in {"GET", "DELETE"}:
            return 1 + max(0, int(cfg.max_retries))

        # Unsafe methods without idempotency → NEVER retry
        if method_u in {"POST", "PATCH", "PUT"} and not self._has_idempotency_key(
            json=json, headers=headers
        ):
            return 1

        # Unsafe but idempotent → retries allowed
        return 1 + max(0, int(cfg.max_retries))


    def close(self) -> None:
        self._client.close()

    def _request_id(self, r: httpx.Response) -> Optional[str]:
        return r.headers.get("x-request-id") or r.headers.get("cf-ray")

    def _parse_json_best_effort(self, r: httpx.Response) -> Any:
        try:
            return r.json()
        except Exception:
            return None

    def _handle_response(self, *, method: str, path: str, r: httpx.Response) -> Any:
        request_id = self._request_id(r)

        if r.status_code in (401, 403):
            raise AuthenticationError(f"Authentication failed (status {r.status_code}).")

        if r.status_code == 402:
            data = self._parse_json_best_effort(r)
            msg = data.get("detail") if isinstance(data, dict) else None
            raise BillingError(msg or "Billing is required for this operation.")

        if r.status_code in (400, 422):
            data = self._parse_json_best_effort(r)
            msg = data.get("detail") if isinstance(data, dict) else None
            raise ValidationError(msg or "Invalid request.")

        if r.status_code == 429:
            raise RateLimitError("Rate limit exceeded.")

        if r.status_code >= 400:
            data = self._parse_json_best_effort(r)
            msg = None
            if isinstance(data, dict):
                msg = data.get("detail") or data.get("message")
            if not msg:
                msg = r.text.strip() or f"HTTP {r.status_code}"
            raise ApiError(
                status_code=r.status_code,
                message=str(msg),
                method=method.upper(),
                path=path,
                response_json=data if isinstance(data, dict) else None,
                request_id=request_id,
            )

        if r.status_code == 204:
            return None

        ctype = (r.headers.get("content-type") or "").lower()
        if "application/json" in ctype:
            return r.json() if r.content else None
        return r.text

    def _retry_after_seconds(self, r: httpx.Response) -> Optional[float]:
        ra = r.headers.get("retry-after")
        if not ra:
            return None
        try:
            return float(ra)
        except Exception:
            return None

    def _backoff(self, attempt_index: int) -> float:
        raw = self._retry.base_backoff_s * (2**attempt_index)
        raw = min(raw, self._retry.max_backoff_s)
        return raw + random.uniform(0.0, self._retry.jitter_s)

    def _should_retry_status(self, status_code: int) -> bool:
        if status_code in (408, 409, 425, 429):
            return True
        return 500 <= status_code <= 599

    def request(
        self,
        method: str,
        path: str,
        *,
        params: Optional[Dict[str, Any]] = None,
        json: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, str]] = None,
        retry: RetryConfig | None = None,
    ) -> Any:
        cfg = retry or self._retry
        method_u = method.upper()
        attempts = self._effective_attempts(
            method_u=method_u,
            json=json,
            headers=headers,
            cfg=cfg,
        )
        
        if self._debug:
            safe_headers = _redact_headers(dict(self._client.headers))
            self._logger.debug(
                "Lamen request: %s %s params=%s json=%s headers=%s",
                method_u,
                path,
                params,
                json,
                safe_headers,
            )

        for i in range(attempts):
            try:
                r = self._client.request(method_u, path, params=params, json=json, headers=headers)
            except httpx.TimeoutException as e:
                if i >= attempts - 1:
                    raise TimeoutError("Request timed out.") from e
                time_s = self._backoff(i)
                import time
                time.sleep(time_s)
                continue
            except httpx.RequestError as e:
                if i >= attempts - 1:
                    raise NetworkError("Network error while calling Lamen API.") from e
                time_s = self._backoff(i)
                import time
                time.sleep(time_s)
                continue

            if i < attempts - 1 and self._should_retry_status(r.status_code):
                ra = self._retry_after_seconds(r)
                time_s = ra if ra is not None else self._backoff(i)
                import time
                time.sleep(time_s)
                continue

            if self._debug:
                self._logger.debug(
                    "Lamen response: %s %s status=%s request_id=%s",
                    method_u,
                    path,
                    r.status_code,
                    self._request_id(r),
                )


            return self._handle_response(method=method_u, path=path, r=r)

        raise NetworkError("Unknown error while calling Lamen API.")

    def get(self, path: str, *, params: Optional[Dict[str, Any]] = None) -> Any:
        return self.request("GET", path, params=params)

    def post(self, path: str, *, json: Optional[Dict[str, Any]] = None) -> Any:
        return self.request("POST", path, json=json)

    def patch(self, path: str, *, json: Optional[Dict[str, Any]] = None) -> Any:
        return self.request("PATCH", path, json=json)

    def delete(self, path: str, *, params: Optional[Dict[str, Any]] = None) -> Any:
        return self.request("DELETE", path, params=params)


class AsyncHttpClient:
    """Low-level async HTTP wrapper with retries and typed errors."""

    def __init__(
        self,
        *,
        api_key: Optional[str] = None,
        base_url: str = DEFAULT_BASE_URL,
        dashboard_token: Optional[str] = None,
        timeout_s: float = DEFAULT_TIMEOUT_S,
        user_agent: str = f"lamen-python/{__version__}",
        retry: RetryConfig | None = None,
        extra_headers: Optional[Dict[str, str]] = None,
        logger: logging.Logger | None = None,
        debug: bool = False,
    ) -> None:
        self._retry = retry or RetryConfig()
        self._logger = logger or logging.getLogger("lamen")
        self._debug = debug

        headers: Dict[str, str] = {
            "User-Agent": user_agent,
            "Accept": "application/json",
        }
        if api_key:
            headers["x-api-key"] = api_key
        if dashboard_token:
            headers["Authorization"] = f"Bearer {dashboard_token}"
        if extra_headers:
            headers.update({k: v for k, v in extra_headers.items() if v is not None})

        self._client = httpx.AsyncClient(
            base_url=base_url.rstrip("/"),
            headers=headers,
            timeout=httpx.Timeout(timeout_s),
        )

    async def aclose(self) -> None:
        await self._client.aclose()

    def _request_id(self, r: httpx.Response) -> Optional[str]:
        return r.headers.get("x-request-id") or r.headers.get("cf-ray")

    def _parse_json_best_effort(self, r: httpx.Response) -> Any:
        try:
            return r.json()
        except Exception:
            return None
        

    def _has_idempotency_key(
        self,
        *,
        json: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, str]] = None,
    ) -> bool:
        if headers:
            for k in ("Idempotency-Key", "X-Idempotency-Key", "x-idempotency-key"):
                v = headers.get(k)
                if v is not None and str(v).strip():
                    return True

        if json and isinstance(json, dict):
            v = json.get("idempotency_key")
            if v is not None and str(v).strip():
                return True

        return False


    def _effective_attempts(
        self,
        *,
        method_u: str,
        json: Optional[Dict[str, Any]],
        headers: Optional[Dict[str, str]],
        cfg: RetryConfig,
    ) -> int:
        if method_u in {"GET", "DELETE"}:
            return 1 + max(0, int(cfg.max_retries))

        if method_u in {"POST", "PATCH", "PUT"} and not self._has_idempotency_key(
            json=json, headers=headers
        ):
            return 1

        return 1 + max(0, int(cfg.max_retries))

    def _handle_response(self, *, method: str, path: str, r: httpx.Response) -> Any:
        request_id = self._request_id(r)

        if r.status_code in (401, 403):
            raise AuthenticationError(f"Authentication failed (status {r.status_code}).")

        if r.status_code == 402:
            data = self._parse_json_best_effort(r)
            msg = data.get("detail") if isinstance(data, dict) else None
            raise BillingError(msg or "Billing is required for this operation.")

        if r.status_code in (400, 422):
            data = self._parse_json_best_effort(r)
            msg = data.get("detail") if isinstance(data, dict) else None
            raise ValidationError(msg or "Invalid request.")

        if r.status_code == 429:
            raise RateLimitError("Rate limit exceeded.")

        if r.status_code >= 400:
            data = self._parse_json_best_effort(r)
            msg = None
            if isinstance(data, dict):
                msg = data.get("detail") or data.get("message")
            if not msg:
                msg = r.text.strip() or f"HTTP {r.status_code}"
            raise ApiError(
                status_code=r.status_code,
                message=str(msg),
                method=method.upper(),
                path=path,
                response_json=data if isinstance(data, dict) else None,
                request_id=request_id,
            )

        if r.status_code == 204:
            return None

        ctype = (r.headers.get("content-type") or "").lower()
        if "application/json" in ctype:
            return r.json() if r.content else None
        return r.text

    def _retry_after_seconds(self, r: httpx.Response) -> Optional[float]:
        ra = r.headers.get("retry-after")
        if not ra:
            return None
        try:
            return float(ra)
        except Exception:
            return None

    def _backoff(self, attempt_index: int) -> float:
        raw = self._retry.base_backoff_s * (2**attempt_index)
        raw = min(raw, self._retry.max_backoff_s)
        return raw + random.uniform(0.0, self._retry.jitter_s)

    def _should_retry_status(self, status_code: int) -> bool:
        if status_code in (408, 409, 425, 429):
            return True
        return 500 <= status_code <= 599

    async def request(
        self,
        method: str,
        path: str,
        *,
        params: Optional[Dict[str, Any]] = None,
        json: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, str]] = None,
        retry: RetryConfig | None = None,
    ) -> Any:
        cfg = retry or self._retry
        method_u = method.upper()
        attempts = self._effective_attempts(
            method_u=method_u,
            json=json,
            headers=headers,
            cfg=cfg,
        )

        for i in range(attempts):
            try:
                r = await self._client.request(method_u, path, params=params, json=json, headers=headers)
            except httpx.TimeoutException as e:
                if i >= attempts - 1:
                    raise TimeoutError("Request timed out.") from e
                await asyncio.sleep(self._backoff(i))
                continue
            except httpx.RequestError as e:
                if i >= attempts - 1:
                    raise NetworkError("Network error while calling Lamen API.") from e
                await asyncio.sleep(self._backoff(i))
                continue

            if i < attempts - 1 and self._should_retry_status(r.status_code):
                ra = self._retry_after_seconds(r)
                await asyncio.sleep(ra if ra is not None else self._backoff(i))
                continue

            return self._handle_response(method=method_u, path=path, r=r)

        raise NetworkError("Unknown error while calling Lamen API.")

    async def get(self, path: str, *, params: Optional[Dict[str, Any]] = None) -> Any:
        return await self.request("GET", path, params=params)

    async def post(self, path: str, *, json: Optional[Dict[str, Any]] = None) -> Any:
        return await self.request("POST", path, json=json)

    async def patch(self, path: str, *, json: Optional[Dict[str, Any]] = None) -> Any:
        return await self.request("PATCH", path, json=json)

    async def delete(self, path: str, *, params: Optional[Dict[str, Any]] = None) -> Any:
        return await self.request("DELETE", path, params=params)



from __future__ import annotations

from typing import Any, Dict, Optional

from .http import HttpClient, AsyncHttpClient
from .types import (
    EventIngestRequest,
    EventIngestResponse,
    RegisterDeviceIdentityRequest,
    DeviceIdentityResponse,
    UnregisterDeviceIdentityRequest,
    UnregisterDeviceIdentityResponse,
    RegisterEmailIdentityRequest,
    EmailIdentityResponse,
    StartEmailVerifyRequest,
    StartEmailVerifyResponse,
    ConfirmEmailVerifyRequest,
    ConfirmEmailVerifyResponse,
    WebTokenRegisterRequest,
    WebTokenUnregisterRequest,
    DirectSendRequest,
    DirectSendResponse,
    BulkEmailImportRequest
)


class MessagingClient:
    """
    Messaging API (API key)
    Base path: /api/messaging

    IMPORTANT:
    - SDK mirrors ONLY public API schemas.
    - No templates, channels, or routing params here.
    - Templates + channels are attached to RULES via dashboard.
    """

    def __init__(self, http: HttpClient):
        self._http = http

    # ----------------------------
    # Events
    # ----------------------------

    def post_event(
        self,
        *,
        event: str,
        external_user_id: Optional[str] = None,
        data: Optional[Dict[str, Any]] = None,
        idempotency_key: Optional[str] = None,
    ) -> EventIngestResponse:
        """
        Ingest a messaging event.

        Matches:
        POST /api/messaging/events
        EventIngestRequest
        """
        payload = EventIngestRequest(
            event=event,
            external_user_id=external_user_id,
            data=data,
            idempotency_key=idempotency_key,
        )
        resp = self._http.post(
            "/api/messaging/events",
            json=payload.model_dump(exclude_none=True),
        )
        return EventIngestResponse.model_validate(resp)

    # ----------------------------
    # Device identities (push)
    # ----------------------------
    
    def send(
        self,
        *,
        channel: str,
        external_user_id: str,
        title: str | None = None,
        body: str | None = None,
        data: dict | None = None,
        priority: str = "standard",
        idempotency_key: str | None = None,
    ) -> DirectSendResponse:
        payload = DirectSendRequest(
            channel=channel,
            external_user_id=external_user_id,
            title=title,
            body=body,
            data=data,
            priority=priority,
            idempotency_key=idempotency_key,
        )

        resp = self._http.post(
            "/api/messaging/send",
            json=payload.model_dump(exclude_none=True),
        )
        return DirectSendResponse.model_validate(resp)
    
    def bulk_import_email_identities(
        self,
        *,
        users: list[dict],
    ) -> None:
        payload = BulkEmailImportRequest(users=users)
        self._http.post(
            "/api/messaging/identities/email/bulk_import",
            json=payload.model_dump(),
        )

    def register_device(
        self,
        *,
        external_user_id: str,
        token: str,
        platform: str,
        app_id: Optional[str] = None,
    ) -> DeviceIdentityResponse:
        """
        Register a push device identity.

        POST /api/messaging/devices
        """
        payload = RegisterDeviceIdentityRequest(
            external_user_id=external_user_id,
            token=token,
            platform=platform,
            app_id=app_id,
        )
        resp = self._http.post("/api/messaging/devices", json=payload.model_dump(exclude_none=True))
        return DeviceIdentityResponse.model_validate(resp)

    def unregister_device(
        self,
        *,
        external_user_id: str,
        token: str,
    ) -> UnregisterDeviceIdentityResponse:
        """
        Unregister a push device identity.

        POST /api/messaging/devices/unregister
        """
        payload = UnregisterDeviceIdentityRequest(
            external_user_id=external_user_id,
            token=token,
        )
        resp = self._http.post("/api/messaging/devices/unregister", json=payload.model_dump())
        return UnregisterDeviceIdentityResponse.model_validate(resp)

    # ----------------------------
    # Email identities
    # ----------------------------

    def register_email_identity(
        self,
        *,
        external_user_id: str,
        email: str,
        label: Optional[str] = None,
    ) -> EmailIdentityResponse:
        """
        Register an email identity.

        POST /api/messaging/identities/email
        """
        payload = RegisterEmailIdentityRequest(
            external_user_id=external_user_id,
            email=email,
            label=label,
        )
        resp = self._http.post(
            "/api/messaging/identities/email",
            json=payload.model_dump(exclude_none=True),
        )
        return EmailIdentityResponse.model_validate(resp)

    def start_email_verification(
        self,
        *,
        external_user_id: str,
        email: str,
        label: Optional[str] = None,
    ) -> StartEmailVerifyResponse:
        """
        Start email verification flow.

        POST /api/messaging/identities/email/start_verify
        """
        payload = StartEmailVerifyRequest(
            external_user_id=external_user_id,
            email=email,
            label=label,
        )
        resp = self._http.post(
            "/api/messaging/identities/email/start_verify",
            json=payload.model_dump(exclude_none=True),
        )
        if resp is None:
            return StartEmailVerifyResponse(ok=True)
        return StartEmailVerifyResponse.model_validate(resp)

    def confirm_email_verification(
        self,
        *,
        external_user_id: str,
        email: str,
        code: str,
    ) -> ConfirmEmailVerifyResponse:
        """
        Confirm email verification.

        POST /api/messaging/identities/email/confirm
        """
        payload = ConfirmEmailVerifyRequest(
            external_user_id=external_user_id,
            email=email,
            code=code,
        )
        resp = self._http.post(
            "/api/messaging/identities/email/confirm",
            json=payload.model_dump(),
        )
        if resp is None:
            return ConfirmEmailVerifyResponse(ok=True)
        return ConfirmEmailVerifyResponse.model_validate(resp)

    # ----------------------------
    # Web push tokens
    # ----------------------------

    def register_web_token(
        self,
        *,
        external_user_id: str,
        token: str,
        app_id: Optional[str] = None,
    ) -> None:
        """
        Register a web push token.

        POST /api/messaging/web_tokens/register
        """
        payload = WebTokenRegisterRequest(
            external_user_id=external_user_id,
            token=token,
            app_id=app_id,
        )
        self._http.post(
            "/api/messaging/web_tokens/register",
            json=payload.model_dump(exclude_none=True),
        )

    def unregister_web_token(
        self,
        *,
        external_user_id: str,
        token: str,
    ) -> None:
        """
        Unregister a web push token.

        POST /api/messaging/web_tokens/unregister
        """
        payload = WebTokenUnregisterRequest(
            external_user_id=external_user_id,
            token=token,
        )
        self._http.post(
            "/api/messaging/web_tokens/unregister",
            json=payload.model_dump(),
        )


# =====================================================================
# ASYNC CLIENT
# =====================================================================

class AsyncMessagingClient:
    def __init__(self, http: AsyncHttpClient):
        self._http = http

    async def post_event(
        self,
        *,
        event: str,
        external_user_id: Optional[str] = None,
        data: Optional[Dict[str, Any]] = None,
        idempotency_key: Optional[str] = None,
    ) -> EventIngestResponse:
        payload = EventIngestRequest(
            event=event,
            external_user_id=external_user_id,
            data=data,
            idempotency_key=idempotency_key,
        )
        resp = await self._http.post(
            "/api/messaging/events",
            json=payload.model_dump(exclude_none=True),
        )
        return EventIngestResponse.model_validate(resp)
    

    async def send(
        self,
        *,
        channel: str,
        external_user_id: str,
        title: str | None = None,
        body: str | None = None,
        data: dict | None = None,
        priority: str = "standard",
        idempotency_key: str | None = None,
    ) -> DirectSendResponse:
        payload = DirectSendRequest(
            channel=channel,
            external_user_id=external_user_id,
            title=title,
            body=body,
            data=data,
            priority=priority,
            idempotency_key=idempotency_key,
        )

        resp = await self._http.post(
            "/api/messaging/send",
            json=payload.model_dump(exclude_none=True),
        )
        return DirectSendResponse.model_validate(resp)
    
    async def bulk_import_email_identities(
        self,
        *,
        users: list[dict],
    ) -> None:
        payload = BulkEmailImportRequest(users=users)
        await self._http.post(
            "/api/messaging/identities/email/bulk_import",
            json=payload.model_dump(),
        )

    async def register_device(
        self,
        *,
        external_user_id: str,
        token: str,
        platform: str,
        app_id: Optional[str] = None,
    ) -> DeviceIdentityResponse:
        payload = RegisterDeviceIdentityRequest(
            external_user_id=external_user_id,
            token=token,
            platform=platform,
            app_id=app_id,
        )
        resp = await self._http.post(
            "/api/messaging/devices",
            json=payload.model_dump(exclude_none=True),
        )
        return DeviceIdentityResponse.model_validate(resp)

    async def unregister_device(
        self,
        *,
        external_user_id: str,
        token: str,
    ) -> UnregisterDeviceIdentityResponse:
        payload = UnregisterDeviceIdentityRequest(
            external_user_id=external_user_id,
            token=token,
        )
        resp = await self._http.post(
            "/api/messaging/devices/unregister",
            json=payload.model_dump(),
        )
        return UnregisterDeviceIdentityResponse.model_validate(resp)

    async def register_email_identity(
        self,
        *,
        external_user_id: str,
        email: str,
        label: Optional[str] = None,
    ) -> EmailIdentityResponse:
        payload = RegisterEmailIdentityRequest(
            external_user_id=external_user_id,
            email=email,
            label=label,
        )
        resp = await self._http.post(
            "/api/messaging/identities/email",
            json=payload.model_dump(exclude_none=True),
        )
        return EmailIdentityResponse.model_validate(resp)

    async def start_email_verification(
        self,
        *,
        external_user_id: str,
        email: str,
        label: Optional[str] = None,
    ) -> StartEmailVerifyResponse:
        payload = StartEmailVerifyRequest(
            external_user_id=external_user_id,
            email=email,
            label=label,
        )
        resp = await self._http.post(
            "/api/messaging/identities/email/start_verify",
            json=payload.model_dump(exclude_none=True),
        )
        if resp is None:
            return StartEmailVerifyResponse(ok=True)
        return StartEmailVerifyResponse.model_validate(resp)

    async def confirm_email_verification(
        self,
        *,
        external_user_id: str,
        email: str,
        code: str,
    ) -> ConfirmEmailVerifyResponse:
        payload = ConfirmEmailVerifyRequest(
            external_user_id=external_user_id,
            email=email,
            code=code,
        )
        resp = await self._http.post(
            "/api/messaging/identities/email/confirm",
            json=payload.model_dump(),
        )
        if resp is None:
            return ConfirmEmailVerifyResponse(ok=True)
        return ConfirmEmailVerifyResponse.model_validate(resp)

    async def register_web_token(
        self,
        *,
        external_user_id: str,
        token: str,
        app_id: Optional[str] = None,
    ) -> None:
        payload = WebTokenRegisterRequest(
            external_user_id=external_user_id,
            token=token,
            app_id=app_id,
        )
        await self._http.post(
            "/api/messaging/web_tokens/register",
            json=payload.model_dump(exclude_none=True),
        )

    async def unregister_web_token(
        self,
        *,
        external_user_id: str,
        token: str,
    ) -> None:
        payload = WebTokenUnregisterRequest(
            external_user_id=external_user_id,
            token=token,
        )
        await self._http.post(
            "/api/messaging/web_tokens/unregister",
            json=payload.model_dump(),
        )



from __future__ import annotations

from uuid import UUID

from .http import HttpClient, AsyncHttpClient
from .types import StreamUrlResponse


class StreamClient:
    """Streaming/CDN signed URL API (API key)."""

    def __init__(self, http: HttpClient):
        self._http = http

    def get_stream_url(self, *, object_id: UUID, expiry_in_seconds: int = 3600) -> StreamUrlResponse:
        data = self._http.get(f"/cdn/{object_id}/url", params={"expiry_in_seconds": expiry_in_seconds})
        return StreamUrlResponse.model_validate(data)
    
    def get_playback_url(self, *, object_id: UUID, expiry_in_seconds: int = 3600) -> StreamUrlResponse:
        data = self._http.get(
            f"/cdn/{object_id}/playback-url",
            params={"expiry_in_seconds": expiry_in_seconds},
        )
        return StreamUrlResponse.model_validate(data)


class AsyncStreamClient:
    def __init__(self, http: AsyncHttpClient):
        self._http = http

    async def get_stream_url(self, *, object_id: UUID, expiry_in_seconds: int = 3600) -> StreamUrlResponse:
        data = await self._http.get(f"/cdn/{object_id}/url", params={"expiry_in_seconds": expiry_in_seconds})
        return StreamUrlResponse.model_validate(data)
    
    async def get_playback_url(self, *, object_id: UUID, expiry_in_seconds: int = 3600) -> StreamUrlResponse:
        data = await self._http.get(
            f"/cdn/{object_id}/playback-url",
            params={"expiry_in_seconds": expiry_in_seconds},
        )
        return StreamUrlResponse.model_validate(data)



from __future__ import annotations
from typing import Any, Callable, Dict, Iterator, List, Optional

def paginate(
    fetch_page: Callable[[Optional[str]], Dict[str, Any]],
    *,
    cursor_key: str = "next_cursor",
    items_key: str = "items",
) -> Iterator[dict]:
    cursor: Optional[str] = None
    while True:
        page = fetch_page(cursor)
        items = page.get(items_key) or []
        for item in items:
            yield item
        cursor = page.get(cursor_key)
        if not cursor:
            break



from __future__ import annotations
from typing import Any, Dict, List, Optional

from .http import HttpClient, AsyncHttpClient
from .types import GeocodeResult
import json
import logging

class GeocodeClient:
    """Geocoding client (sync)."""

    def __init__(self, http: HttpClient, *, logger: logging.Logger | None = None):
        self._http = http
        self._logger = logger or logging.getLogger("lamen.geocode")

    def search(
        self,
        *,
        query: str,
        limit: int = 10,
        country: Optional[str] = None,
    ) -> List[Dict[str, Any]]:
        if not query or not query.strip():
            raise ValueError("Query cannot be empty")

        params: Dict[str, Any] = {
            "q": query,
            "limit": limit,
        }
        if country:
            params["country"] = country

        data = self._http.get("/geocode/search", params=params)

        if isinstance(data, dict) and "value" in data:
            data = data["value"]

        if isinstance(data, str):
            data = json.loads(data)

        if not isinstance(data, list):
            raise ValueError("Unexpected geocode response shape")

        # Now it MUST be a list of dicts
        return [GeocodeResult(**item) for item in data]


class AsyncGeocodeClient:
    """Geocoding client (async)."""

    def __init__(self, http: AsyncHttpClient, *, logger: logging.Logger | None = None):
        self._http = http
        self._logger = logger or logging.getLogger("lamen.geocode")

    async def search(
        self,
        *,
        query: str,
        limit: int = 10,
        country: Optional[str] = None,
    ) -> List[Dict[str, Any]]:
        if not query or not query.strip():
            raise ValueError("Query cannot be empty")

        params: Dict[str, Any] = {
            "q": query,
            "limit": limit,
        }
        if country:
            params["country"] = country

        data = await self._http.get("/geocode/search", params=params)

        #  If Redis-wrapped
        if isinstance(data, dict) and "value" in data:
            data = data["value"]

        #  If JSON string
        if isinstance(data, str):
            data = json.loads(data)

        if not isinstance(data, list):
            raise ValueError("Unexpected geocode response shape")

        # Now it MUST be a list of dicts
        return [GeocodeResult(**item) for item in data]




from __future__ import annotations

from .http import DEFAULT_BASE_URL, DEFAULT_TIMEOUT_S, HttpClient, AsyncHttpClient, RetryConfig
from .storage import StorageClient, AsyncStorageClient
from .stream import StreamClient, AsyncStreamClient
from .athena import AthenaClient, AsyncAthenaClient
from .messaging import MessagingClient, AsyncMessagingClient
from .geocoding import GeocodeClient, AsyncGeocodeClient
from ._version import __version__


class Lamen:
    """Top-level SDK client (sync)."""

    def __init__(
        self,
        *,
        api_key: str,
        base_url: str = DEFAULT_BASE_URL,
        timeout_s: float = DEFAULT_TIMEOUT_S,
        user_agent: str = f"lamen-python/{__version__}",
        retry: RetryConfig | None = None,
    ):
        self._http = HttpClient(
            api_key=api_key,
            base_url=base_url,
            timeout_s=timeout_s,
            user_agent=user_agent,
            retry=retry,
        )

        self.storage = StorageClient(self._http)
        self.stream = StreamClient(self._http)
        self.athena = AthenaClient(self._http)
        self.messaging = MessagingClient(self._http)
        self.geocode = GeocodeClient(self._http)

    def close(self) -> None:
        self._http.close()

    def __enter__(self) -> "Lamen":
        return self

    def __exit__(self, exc_type, exc, tb) -> None:
        self.close()


class AsyncLamen:
    """Top-level SDK client (async)."""

    def __init__(
        self,
        *,
        api_key: str,
        base_url: str = DEFAULT_BASE_URL,
        timeout_s: float = DEFAULT_TIMEOUT_S,
        user_agent: str = f"lamen-python/{__version__}",
        retry: RetryConfig | None = None,
    ):
        self._http = AsyncHttpClient(
            api_key=api_key,
            base_url=base_url,
            timeout_s=timeout_s,
            user_agent=user_agent,
            retry=retry,
        )

        self.storage = AsyncStorageClient(self._http)
        self.stream = AsyncStreamClient(self._http)
        self.athena = AsyncAthenaClient(self._http)
        self.messaging = AsyncMessagingClient(self._http)
        self.geocode = AsyncGeocodeClient(self._http) 

    async def aclose(self) -> None:
        await self._http.aclose()

    async def __aenter__(self) -> "AsyncLamen":
        return self

    async def __aexit__(self, exc_type, exc, tb) -> None:
        await self.aclose()



from .client import Lamen, AsyncLamen
from .errors import (
    ApiError,
    AuthenticationError,
    BillingError,
    NetworkError,
    RateLimitError,
    TimeoutError,
    ValidationError,
)
from ._version import __version__


__all__ = [
    "Lamen",
    "AsyncLamen",
    "__version__",
    "ApiError",
    "AuthenticationError",
    "BillingError",
    "NetworkError",
    "RateLimitError",
    "TimeoutError",
    "ValidationError",
]




from __future__ import annotations

from typing import List, Optional
from uuid import UUID

from .http import HttpClient, AsyncHttpClient
from .types import (
    StorageUploadRequest,
    StorageUploadResponse,
    StorageConfirmRequest,
    StorageDownloadRequest,
    StorageDownloadResponse,
    StorageObjectOut,
    StorageDeleteResponse,
    StoragePrivacyRequest,
    FolderCreateRequest,
    FolderOut,
    FolderDeleteResponse,
)

class StorageClient:
    """Storage API (API key)."""
    def __init__(self, http: HttpClient):
        self._http = http

    def create_upload_url(
        self,
        *,
        filename: str,
        content_type: str,
        size_bytes: int,
        folder_id: Optional[UUID] = None,
        is_public: bool = False,
        generate_thumbnail: bool = False,
        transcode_preset: str = "none",
    ) -> StorageUploadResponse:
        payload = StorageUploadRequest(
            filename=filename,
            content_type=content_type,
            size_bytes=size_bytes,
            folder_id=folder_id,
            is_public=is_public,
            generate_thumbnail=generate_thumbnail,
            transcode_preset=transcode_preset,
        )
        data = self._http.post("/storage/upload-url/api", json=payload.model_dump(mode="json", exclude_none=True))
        return StorageUploadResponse.model_validate(data)

    def confirm_upload(self, *, object_id: UUID) -> StorageObjectOut:
        payload = StorageConfirmRequest(object_id=object_id)
        data = self._http.post("/storage/confirm/api", json=payload.model_dump(mode="json"))
        return StorageObjectOut.model_validate(data)

    def create_download_url(self, *, object_id: UUID) -> StorageDownloadResponse:
        payload = StorageDownloadRequest(object_id=object_id)
        data = self._http.post("/storage/download-url/api", json=payload.model_dump(mode="json"))
        return StorageDownloadResponse.model_validate(data)

    def list_objects(self, *, folder_id: Optional[UUID] = None) -> List[StorageObjectOut]:
        params = {"folder_id": str(folder_id)} if folder_id else None
        data = self._http.get("/storage/list/api", params=params)
        return [StorageObjectOut.model_validate(x) for x in (data or [])]

    def set_object_public(self, *, object_id: UUID, is_public: bool) -> StorageObjectOut:
        payload = StoragePrivacyRequest(is_public=is_public)
        data = self._http.patch(f"/storage/object/api/{object_id}/privacy", json=payload.model_dump())
        return StorageObjectOut.model_validate(data)

    def delete_object(self, *, object_id: UUID) -> StorageDeleteResponse:
        data = self._http.delete(f"/storage/object/api/{object_id}")
        return StorageDeleteResponse.model_validate(data)

    def create_folder(self, *, path: str) -> FolderOut:
        payload = FolderCreateRequest(path=path)
        data = self._http.post("/storage/folders/api", json=payload.model_dump())
        return FolderOut.model_validate(data)

    def list_folders(self) -> List[FolderOut]:
        data = self._http.get("/storage/folders/api")
        return [FolderOut.model_validate(x) for x in (data or [])]

    def delete_folder(self, *, folder_id: UUID) -> FolderDeleteResponse:
        data = self._http.delete(f"/storage/folders/api/{folder_id}")
        return FolderDeleteResponse.model_validate(data)


class AsyncStorageClient:
    def __init__(self, http: AsyncHttpClient):
        self._http = http

    async def create_upload_url(
        self,
        *,
        filename: str,
        content_type: str,
        size_bytes: int,
        folder_id: Optional[UUID] = None,
        is_public: bool = False,
        generate_thumbnail: bool = False,
        transcode_preset: str = "none",
    ) -> StorageUploadResponse:
        payload = StorageUploadRequest(
            filename=filename,
            content_type=content_type,
            size_bytes=size_bytes,
            folder_id=folder_id,
            is_public=is_public,
            generate_thumbnail=generate_thumbnail,
            transcode_preset=transcode_preset,
        )
        data = await self._http.post("/storage/upload-url/api", json=payload.model_dump(exclude_none=True))
        return StorageUploadResponse.model_validate(data)

    async def confirm_upload(self, *, object_id: UUID) -> StorageObjectOut:
        payload = StorageConfirmRequest(object_id=object_id)
        data = await self._http.post("/storage/confirm/api", json=payload.model_dump(mode="json"))
        return StorageObjectOut.model_validate(data)

    async def create_download_url(self, *, object_id: UUID) -> StorageDownloadResponse:
        payload = StorageDownloadRequest(object_id=object_id)
        data = await self._http.post("/storage/download-url/api", json=payload.model_dump())
        return StorageDownloadResponse.model_validate(data)

    async def list_objects(self, *, folder_id: Optional[UUID] = None) -> List[StorageObjectOut]:
        params = {"folder_id": str(folder_id)} if folder_id else None
        data = await self._http.get("/storage/list/api", params=params)
        return [StorageObjectOut.model_validate(x) for x in (data or [])]

    async def set_object_public(self, *, object_id: UUID, is_public: bool) -> StorageObjectOut:
        payload = StoragePrivacyRequest(is_public=is_public)
        data = await self._http.patch(f"/storage/object/api/{object_id}/privacy", json=payload.model_dump())
        return StorageObjectOut.model_validate(data)

    async def delete_object(self, *, object_id: UUID) -> StorageDeleteResponse:
        data = await self._http.delete(f"/storage/object/api/{object_id}")
        return StorageDeleteResponse.model_validate(data)

    async def create_folder(self, *, path: str) -> FolderOut:
        payload = FolderCreateRequest(path=path)
        data = await self._http.post("/storage/folders/api", json=payload.model_dump())
        return FolderOut.model_validate(data)

    async def list_folders(self) -> List[FolderOut]:
        data = await self._http.get("/storage/folders/api")
        return [FolderOut.model_validate(x) for x in (data or [])]

    async def delete_folder(self, *, folder_id: UUID) -> FolderDeleteResponse:
        data = await self._http.delete(f"/storage/folders/api/{folder_id}")
        return FolderDeleteResponse.model_validate(data)
